/*!
 * @file fas_trace.h
 * @brief Fastec Tracing support library.
 *
 * Copyright 2015 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef FAS_TRACE_H_
#define FAS_TRACE_H_
#pragma once

#include <stdarg.h>
#include <stdio.h>
#include <stdint.h>

#include <fas_log.h>

extern void trace(const char *fmt, ...);
extern void vtrace(const char *fmt, va_list ap);

/**
 * @brief Use this macro to enable 'global' trace control for a source file
 *
 * This macro forces use of the global (not in a namespace) Trace Control word.
 */
#define TRACE_GLOBAL  extern uint32_t traceControl

/**
 * @brief Use this macro to enable a local (to source file) Trace Control word.
 *
 * This macro defines a local (static) traceControl variable.
 */
#define TRACE_FILE   static uint32_t traceControl = TRACE_DEFAULT

#define TRACE_LEVEL   0x00000003
#define TRACE_0       0x00000000  // Lowest
#define TRACE_1       0x00000001
#define TRACE_2       0x00000002
#define TRACE_3       0x00000003  // Highest

#define TRACE_SEL_DEFAULT 0x00000004

#define TRACE_DEFAULT_MASK (TRACE_SEL_DEFAULT | TRACE_1)

#define TRACE_OFF()     traceControl = 0

/**
 * @brief Default trace point
 *
 * Only selection is the per-file trace control word's TRACE_DEFAULT bit
 *
 * @param fmt printf()-style format string, followed by arguments
 */
#if defined(_WIN32) && !defined(__MINGW32__)
#define TP_DEF(fmt, ...)        do {                                                                \
                                    if (((traceControl & TRACE_LEVEL) <= 0) &&                      \
                                        ((traceControl & TRACE_DEFAULT_MASK & ~TRACE_LEVEL) != 0)) {\
                                        LOG(fmt, __VA_ARGS__ );                                     \
                                   }                                                                \
                                } while(0)
#else
#define TP_DEF(fmt, args...)    do {                                                                 \
                                    if (((traceControl & TRACE_LEVEL) <= 0) &&                       \
                                        ((traceControl & TRACE_DEFAULT_MASK & ~TRACE_LEVEL) != 0)) { \
                                        LOG(fmt, ##args);                                            \
                                    }                                                                \
                                } while(0)
#endif // !(defined(_WIN32) && !defined(__MINGW32__))

/**
 * @brief Mask-qualified trace point
 *
 * Invoked if any of the per-file trace control word bits in the trace Point's
 * mask word are set.
 *
 * @param mask Mask word for this Trace point
 * @param fmt  printf()-style format string, followed by arguments
 */
#if defined(_WIN32) && !defined(__MINGW32__)
#define TP(mask, fmt, ...)      do {                                                              \
                                    if (((traceControl & TRACE_LEVEL) <= (mask & TRACE_LEVEL)) &&  \
                                        ((traceControl & mask & ~TRACE_LEVEL) != 0)) {             \
                                        LOG(fmt, __VA_ARGS__);                                     \
                                    }                                                              \
                                } while(0)
#else
#define TP(mask, fmt, args...)  do {                                                              \
                                    if (((traceControl & TRACE_LEVEL) <= (mask & TRACE_LEVEL)) && \
                                        ((traceControl & mask & ~TRACE_LEVEL) != 0)) {            \
                                        LOG(fmt, ##args);                                         \
                                    }                                                             \
                                } while(0)
#endif // !(defined(_WIN32) && !defined(__MINGW32__))

/**
* @brief Trace base class
*
* Any class can inherit from this base class to implement an instance-specific
* traceControl word. The base class provides set/get methods to control tracing
* within each derived class instance.
*/
class Trace
{
public:
    Trace();
    virtual ~Trace();

    void setTrace(uint32_t t);
    uint32_t getTrace();

protected:
    uint32_t traceControl;
};

#endif /* FAS_TRACE_H_ */
