/*!
 * @file ffmpeg_encoder.h
 * @brief Interface for FFmpeg decoder object.
 *
 * Copyright 2020 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef FFMPEG_ENCODER_H_
#define FFMPEG_ENCODER_H_
#pragma once

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <memory>
#include <mutex>

#ifdef  __cplusplus
extern "C" {
#endif

#include "libavformat/avformat.h"
#include "libavfilter/buffersink.h"
#include "libavfilter/buffersrc.h"
#include "libavformat/avio.h"
#include "libavcodec/avcodec.h"
#include "libavutil/file.h"
#include "libavutil/error.h"
#include "libavutil/opt.h"
#include "libavutil/pixdesc.h"

#ifdef  __cplusplus
}
#endif

#include "fas_video.h"
#include "fi_durations.h"
#include "ffmpeg_container.h"

namespace FI_FFMPEG
{
//--------------------------------------------------------------------
// Constants, Typedefs and Macros
//--------------------------------------------------------------------

//! These status conditions define operational status of the frame writer thread.
typedef enum FRAME_WRITER_STATUS
{
    ALL_OK = 0,
    TOO_MANY_LOST_FRAMES,
    NO_FRAME_TO_DUPLICATE,
    INVALID_ENCODER_FRAME_TYPE,
    MISSING_FLUSH_FRAME
} FRAME_WRITER_STATUS;

//! The frame writer status is maintained in this record stored as a private data
//! record in the FFmpegEncoder object and updated by the frame writer thread when
//! frame data is processed.
typedef struct FRAME_WRITER_STATUS_REC
{
    uint32_t framesEncoded;                         //!< How many frames encoded in frame writer thread so far.
    bool ffmpegEOF;                                 //!< Was an EOF condition hit?
    int ffmpegError;                                //!< Was an FFmpeg Error code returned?
    FRAME_WRITER_STATUS frameWriterThreadStatus;    //!< The frame writer thread's status state.
    bool wroteAVTrailer;                            //!< Was the AVTrailer written to the stream?
} FRAME_WRITER_STATUS_REC;

//--------------------------------------------------------------------
// Class Declarations
//--------------------------------------------------------------------

class FFmpegFilters;    // Forward class definition.

//! Create a class to handle managment of the encoder information.
class FFmpegEncoder
{
public:
    virtual ~FFmpegEncoder();

    static void CloseEncoderInstance();

    static std::shared_ptr<FFmpegEncoder> CreateEncoder(CONTAINER_TYPE containerType,
                                                        const std::string& videoName,
                                                        enum AVPixelFormat avHWDeviceToUseIfAvailable,
                                                        enum AVPixelFormat avInPixelFormatToUse,
                                                        enum AVPixelFormat avOutPixelFormatToUse,
                                                        uint32_t frameRateToUse,
                                                        uint32_t frameCountToUse,
                                                        uint32_t frameWidthToUse,
                                                        uint32_t frameHeightToUse,
                                                        uint32_t crfValue,
                                                        const std::string& presetValue,
                                                        uint32_t keyFrameIntervalToUse);

    static std::shared_ptr<FFmpegEncoder> GetEncoderInstance();

    static bool IsAMFAvailable();
    static bool IsNVIDIAAvailable();
    static bool IsQSVAvailable();
    static bool IsVAAPIAvailable();

    bool IsPrepared();
    bool IsComplete(bool& inError);

    int EncodeAVFrame(AVFrame **frame);
    bool SendFlushEncoderFrame();

    AVFormatContext *GetAVFormatContext();
    AVCodecContext *GetAVCodecContext();

    enum AVPixelFormat GetPixelInputFormat();
    enum AVPixelFormat GetPixelOutputFormat();
    enum AVPixelFormat GetPixelHWFormat();

    bool GetAspectRatio(AVRational& avrAspectRatioToSet);
    bool GetTimeBase(AVRational& avrTimeBaseToSet);
    bool GetFrameRate(AVRational& avrFrameRateToSet);
    uint32_t GetFrameCount();
    uint32_t GetFrameWidth();
    uint32_t GetFrameHeight();
    uint32_t GetEncodedFrameCount(bool& isPrepared, bool& isComplete, bool& hasError);

    void LogStatistics();

private:
    // This is a singleton object created by the factory method CreateEncoder.
    FFmpegEncoder() :
        fmt_ctx(nullptr),
        stream(nullptr),
        codec(nullptr),
        codec_ctx(nullptr),
        avHWDevicePixelFormat(AV_PIX_FMT_NONE),
        avInPixelFormat(AV_PIX_FMT_NONE),
        avOutPixelFormat(AV_PIX_FMT_NONE),
        avrAspectRatio{1, 1},
        avrTimeBase{1, FAS_VIDEO::FI_DEFAULT_FRAME_RATE},
        avrFrameRate{FAS_VIDEO::FI_DEFAULT_FRAME_RATE, 1},
        frameWidth(0),
        frameHeight(0),
        frameCount(0),
        framesRead(0),
        ffmpegStatusFromCreation(0),
        encoderPrepared(false),
        encodeAVFrameTimeout(false),
        filterPtr(nullptr),
        keyFrameInterval(FAS_VIDEO::DEFAULT_H264_KEYFRAME_INTERVAL)
    {
        theFrameWriterStatus.framesEncoded = 0;
        theFrameWriterStatus.ffmpegEOF = false;
        theFrameWriterStatus.ffmpegError = 0;
        theFrameWriterStatus.frameWriterThreadStatus = ALL_OK;
        theFrameWriterStatus.wroteAVTrailer = false;
    }

    // There is only one me in this world.
    static std::shared_ptr<FFmpegEncoder> FFmpegEncoderInstance;

    // For hardware acceleration there is only the one context.
    static AVBufferRef *HWDeviceCTX;

    //! FFmpeg specific data structures necessary for encoding frame data.
    AVFormatContext *fmt_ctx;                   //!< The FFmpeg AVFormatContext record.
    AVStream *stream;                           //!< The FFmpeg AVStream record.
    const AVCodec *codec;                       //!< The FFmpeg AVCodec record.
    AVCodecContext *codec_ctx;                  //!< The FFmpeg AVCodecContext context record.

    enum AVPixelFormat avHWDevicePixelFormat;   //!< If HW Encoding this is the type of HW Encoder.
    enum AVPixelFormat avInPixelFormat;         //!< The FFmpeg AV pixel format of input frames to encoder.
    enum AVPixelFormat avOutPixelFormat;        //!< The FFmpeg AV pixel format of output frames from encoder.

    AVRational avrAspectRatio;                  //!< The aspect ratio used for encoder.
    AVRational avrTimeBase;                     //!< The default time base used for encoder.
    AVRational avrFrameRate;                    //!< The default time base used for encoder.

    uint32_t frameWidth;                        //!< How many pixels across are the frames?
    uint32_t frameHeight;                       //!< How many rows are in a frame?

    uint32_t frameCount;                        //!< The number of frames expected to encode.
    uint32_t framesRead;                        //!< The number of frames received from stream for encoding.

    int ffmpegStatusFromCreation;               //!< Persist the ffmpeg status code when encoder was created.
    bool encoderPrepared;                       //!< True if ready to encoder, false if not.
    bool encodeAVFrameTimeout;                  //!< If True a timeout occurred in backpressure handler for EncodeAVFrame.

    std::shared_ptr<FFmpegFilters> filterPtr;   //!< The FFmpeg Filter object for this container.
    std::shared_ptr<FAS_VIDEO::FiDurations> encoderStats;
    std::unique_ptr<FAS_VIDEO::FiDurations> encodeAVFrameStats;

    uint32_t keyFrameInterval;              //!< The keyframe interval used when encoding video stream.

    FRAME_WRITER_STATUS_REC theFrameWriterStatus;    //!< One record identifies status for frame writer.

    bool PollFrameWriterThreadStatus();

    AVFrame *CreateFilteredAVFrame(AVFrame *sourceFrame, int& ffmpegResultRef);
    AVFrame *CreateHWAVFrame(AVFrame *sourceFrame, int& ffmpegResultRef);

    static int PrepareHWDeviceContext(AVCodecContext *ctx, AVBufferRef *ourHWDeviceCtx, AVPixelFormat swAVPixFmtOfFrame);
};

} /* End of namespace FI_FFMPEG */

#endif /* !FFMPEG_ENCODER_H_ */
