/*!
 * @file ffmpeg_status.h
 * @brief Provides status support object.
 *
 * Copyright 2019 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef FFMPEG_STATUS_H
#define FFMPEG_STATUS_H

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <stdint.h>
#include <string>

//! @brief Translates FFmpeg error codes into user readable error string and logs it.
//!
//! @param av_code  The FFmpeg error code returned from FFmpeg library call.
//!
extern void log_ffmpeg_error_str(int av_code);

namespace FI_FFMPEG
{
//--------------------------------------------------------------------
// Constants, Typedefs and Macros
//--------------------------------------------------------------------

//! Create object conditions of status for a FFmpeg container context.
typedef enum CTX_STATUS
{
    CONSTRUCTED,            //!< An FI_FFMPEG container context is only constructed and not ready to be used.
    PREPARED,               //!< An FI_FFMPEG container context is prepared and ready for use.
    BAD_ALLOC,              //!< An FI_FFMPEG container context was unable to allocate internal objects.
    BAD_EXCEPT,             //!< An FI_FFMPEG container context received unhandled exception from system.
    BAD_FILE,               //!< An FI_FFMPEG container context received file information for non-existant file.
    NO_FILE_CTRL,           //!< An FI_FFMPEG container context can't use or is missing file information.
    NO_FMT_CTX,             //!< An FI_FFMPEG container context unable to find valid format context for I/O.
    NO_STREAM,              //!< An FI_FFMPEG container context is missing an I/O video stream.
    NO_CODEC,               //!< An FI_FFMPEG container context unable to find valid codec for I/O.
    NO_IMG_INFO,            //!< An FI_FFMPEG container context unable to find image information for I/O.
    NO_CTX_TYPE,            //!< An FI_FFMPEG container context unable to determine container type.
    NO_CTX_OPS,             //!< An FI_FFMPEG container context unable to determine container context operation.
    NO_CTX_DECODER,         //!< An FI_FFMPEG container context is missing a decoder for transcoding.
    NO_CTX_FILTERS,         //!< An FI_FFMPEG container context is missing a filter for encoding.
    NO_CTX_ENCODER,         //!< An FI_FFMPEG container context is missing a encoder for encoding.
    NO_CTX_FFMPEG_IO,       //!< An FI_FFMPEG container context is missing a FFpeg I/O object.
    ENCODER_TIMEOUT,        //!< The Encoder hit timeout threshold while encoding AVFrames.
    FFMPEG_ERR,             //!< An FI_FFMPEG container context encountered FFmpeg Library error and can't continue.
} CTX_STATUS;

//--------------------------------------------------------------------
// Function Prototypes
//--------------------------------------------------------------------


//--------------------------------------------------------------------
// Class Declarations
//--------------------------------------------------------------------

//! FFMPEG general status object responsible for managing types of status for this namespace collection.
class FFmpegStatus
{
public:
    FFmpegStatus();
    virtual ~FFmpegStatus();

    //! @{
    //! Provide status query methods.
    bool isConstructed();
    bool isPrepared();
    void get_status_msg(std::string& statusMsg);
    //! @}

    //! @{
    //! Provide status set methods.
    void set_prepared(bool log_it = true, const char *more_info = nullptr);
    void set_bad_alloc(bool log_it = true, const char *more_info = nullptr);
    void set_bad_except(bool log_it = true, const char *more_info = nullptr);
    void set_bad_file(bool log_it = true, const char *more_info = nullptr);
    void set_no_file_info(bool log_it = true, const char *more_info = nullptr);
    void set_no_fmt_ctx(bool log_it = true, const char *more_info = nullptr);
    void set_no_stream(bool log_it = true, const char *more_info = nullptr);
    void set_no_codec(bool log_it = true, const char *more_info = nullptr);
    void set_no_img_info(bool log_it = true, const char *more_info = nullptr);
    void set_no_ctx_type(bool log_it = true, const char *more_info = nullptr);
    void set_no_ctx_ops(bool log_it = true, const char *more_info = nullptr);
    void set_no_ctx_decoder(bool log_it = true, const char *more_info = nullptr);
    void set_no_ctx_filters(bool log_it = true, const char *more_info = nullptr);
    void set_no_ctx_encoder(bool log_it = true, const char *more_info = nullptr);
    void set_no_ctx_ffmpeg_io(bool log_it = true, const char *more_info = nullptr);
    void set_encoder_timeout(bool log_it = true, const char *more_info = nullptr);

    void set_ffmpeg_err(int ffmpeg_error, bool log_it = true, const char *more_info = nullptr);
    //! @}

    //! @{
    //! Provide status logging methods.
    void logStatus(const char *addToLogMessage = nullptr);
    //! @}

protected:
    CTX_STATUS status;           //!< The status of I/O for the ffmpegContainer Hierarchy.
    int ffmpeg_result;          //!< Holds the ffmpeg error code if status = FFMPEG_ERR.
};


} // End of FI_FFMPEG namespace.

#endif // FFMPEG_STATUS_H
