/*!
 * @file fi_exception.h
 * @brief Definitions and declarations for the Fastec Exception
 *    used by the Fastec Thread library.
 *
 * Copyright 2019 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef FI_EXCEPTION_H_
#define FI_EXCEPTION_H_
#pragma once

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#if defined(_WIN32) && !defined(__MINGW32__)
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <strsafe.h>
#endif // _WIN32

#include <stdexcept>
#include <exception>

#include "fi_note.h"

namespace FAS_THREAD
{
//--------------------------------------------------------------------
// Constant Definitions
//--------------------------------------------------------------------

/*!
 * @brief Create, but do not throw, a new Logic Error exception.
 *
 * Creates a new exception, with a one-line Note that includes the creation
 * timestamp, file and line number tags, and caller's printf()-formatted
 * one-line message. Additional lines may be appended with the FI_EXC_APPEND()
 * macro.
 */
#define NEW_LOGIC_ERROR(e, ...) FAS_THREAD::LogicError e(__FILE__, __LINE__, __VA_ARGS__)

/*!
 * @brief Throw a "logic error" exception
 *
 * This macro creates and throws a "logic error" exception, with a formatted
 * message tagged with the source file and line number at which the macro was
 * invoked. The variable-length argument list must start with a printf()-style
 * format string, followed by zero or more arguments (as required by the format
 * strings conversion specifiers).
 *
 * Note that macro includes the 'throw' clause. Usage:
 *
 *      ....computation reveals an invalid value in variable "x":
 *      FI_LOGIC_ERROR("Bug here. x == %d", x);
 *
 * Do not include a newline at the end of your format string.
 */
#define FI_LOGIC_ERROR(...)     do {                                            \
                                    FAS_THREAD::LogicError e;                   \
                                    e.append(__FILE__, __LINE__, __VA_ARGS__);  \
                                    throw e;                                    \
                                } while (0)

/*!
 * @brief Throw a "runtime error" exception
 *
 * This macro creates and throws a "runtime error" exception, with a formatted
 * message tagged with the source file and line number at which the macro was
 * invoked. The variable-length argument list must start with a printf()-style
 * format string, followed by zero or more arguments (as required by the format
 * strings conversion specifiers).
 *
 * Note that macro includes the 'throw' clause. Usage:
 *
 *      ....computation reveals an I/O fault (possible HW failure perhaps):
 *      FI_LOGIC_ERROR("I/O failed: %s", strerror(errno));
 *
 * Do not include a newline at the end of your format string.
 */
#define FI_RUNTIME_ERROR(...)   do {                                            \
                                    FAS_THREAD::RuntimeError e;                 \
                                    e.append(__FILE__, __LINE__, __VA_ARGS__);  \
                                    throw e;                                    \
                                } while (0)

/*!
 * @brief Throw a runtime error exception, with errno string
 *
 * This is a shorthand macro to include errno information in an exception
 * message. An errno value must be supplied as the first macro argument,
 * and a printf()-style format string must be supplied as the second,
 * followed by zero or more arguments for formatted output.
 *
 * The macro automatically adds a line to the exception message with the
 * verbal interpretation of the supplied errno. See strerror() library
 * function.
 */
#if defined(_WIN32) && !defined(__MINGW32__)
#define FI_OS_ERROR(no, ...)    do {                                                \
                                    LPVOID lpMsgBuf;                                \
                                    DWORD _no = no;                                 \
                                    FormatMessage(                                  \
                                        FORMAT_MESSAGE_ALLOCATE_BUFFER |            \
                                        FORMAT_MESSAGE_FROM_SYSTEM     |            \
                                        FORMAT_MESSAGE_IGNORE_INSERTS,              \
                                        NULL,                                       \
                                        _no,                                        \
                                        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),  \
                                        (LPTSTR)&lpMsgBuf,                          \
                                        0, NULL );                                  \
                                    FAS_THREAD::RuntimeError e;                     \
                                    e.append(__FILE__, __LINE__, __VA_ARGS__);      \
                                    e.append(__FILE__, __LINE__, "Errno indicates \"%s\"", (char*)lpMsgBuf); \
                                    throw e;                                        \
                                } while (0)
#else
#define FI_OS_ERROR(no, ...)    do {                                                \
                                    int _no = no;                                   \
                                    FAS_THREAD::RuntimeError e;                     \
                                    e.append(__FILE__, __LINE__, __VA_ARGS__);      \
                                    e.append(__FILE__, __LINE__, "Errno indicates \"%s\"", std::strerror(_no)); \
                                    throw e;                                        \
                                } while (0)
#endif // !( defined(_WIN32) && !defined(__MINGW32__))

/*!
 * @brief Append a line to an exception's message
 *
 * This macro appends a formatted line to the message in an exception. The
 * formatted line is prepended with the source file name and line number
 * at which the macro was invoked. The variable-length argument list must
 * start with a printf()-style format string, followed by zero or more
 * arguments (as required by the format strings conversion specifiers).
 *
 * This is intended for cases where an exception handler will re-throw a
 * caught exception, but can add useful context to the exception's message.
 * Something like this:
 *
 *      catch (LogicError& e) {
 *          FI_EXC_APPEND("Exception during initialization phase %d\n", phase);
 *          throw e;
 *      }
 *
 * Do not include a newline at the end of your format string.
 */
#define FI_EXC_APPEND(e, ...)       e.append(__FILE__, __LINE__, __VA_ARGS__)

//--------------------------------------------------------------------
// Forward Declarations
//--------------------------------------------------------------------

//--------------------------------------------------------------------
// Class Declaration
//--------------------------------------------------------------------

class Exception : public Note
{
public:
    Exception();
    Exception(const Exception& e);
    Exception(const char *f, int n, const char *fmt, ...);

    virtual ~Exception() {}

    void append(const char *fmt, ...);
    virtual void append(const char *file, int line, const char *fmt, ...);
};

class LogicError : public Exception
{
public:
    LogicError();
    LogicError(const LogicError& e);
    LogicError(const char *f, int n, const char *fmt, ...);
};

/*!
 * @brief Fastec Imaging "runtime error" exception class.
 *
 * This is the sub-class of FAS_THREAD::Exception intended to denote fault conditions
 * detected at runtime that are not due to logic flaws in the software.
 * These might be errors returned by the operating system, for example.
 *
 * This class does not add any functionality to the underlying base
 * class. It is declared only to make a distinct type that can be
 * distinguished in catch() clauses.
 */
class RuntimeError : public Exception { };

} /* namespace FAS_THREAD */

#endif /* FI_EXCEPTION_H_ */
