/*!
 * @file fi_image.h
 * @brief The Fastec Image Class.
 *
 * Copyright 2019 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef FASTEC_IMAGE_H_
#define FASTEC_IMAGE_H_

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <stdint.h>
#include <memory>

#ifdef _WIN32
#include "gige_msg.h"
#else
#include <fastec/gige_msg.h>
#endif

#include <fas_video.h>
#include <fi_frame.h>

namespace FAS_VIDEO
{
//--------------------------------------------------------------------
// Constant Definitions
//--------------------------------------------------------------------

//! Defines the image formats we manage with an FiImage object.
typedef enum
{
    //! @name Image formats we support.
    //! @{
    FI_JPEG_IMAGE,              //!< Compressed JPEG image format.
    FI_BMP_IMAGE,               //!< Bitmap image file or device independent bitmap (DIB)
    FI_DNG_IMAGE,               //!< Adobe Digital Negative Raw Image.
    FI_TIF_IMAGE,               //!< Tagged Image File Format Image.
    FI_BMP_PLANAR_IMAGE,        //!< Bitmap planar image file for Y8 and NV12 image data.
    FI_RDI_PACKED_IMAGE,        //!< RDI packed image file format (proprietary to RDI).
    //! @}

    FI_INVALID_IMAGE            //!< Oops! This one isn't allowed!
} IMAGE_FORMAT;

//! Create object conditions for status of FiImage.
typedef enum IMAGE_STATUS
{
    IMAGE_NEW,              //!< FiImage object is only constructed and not ready to be used.
    IMAGE_EMPTY,            //!< FiImage object has had image attached but is now empty.
    IMAGE_ATTACHED,         //!< FiImage object is prepared and ready for use.
    BAD_IMAGE_FORMAT        //!< FiImage object could not validate image format.
} IMAGE_STATUS;

//--------------------------------------------------------------------
// Forward Declarations
//--------------------------------------------------------------------
class FiImagePkg;

//--------------------------------------------------------------------
// Class Declaration
//--------------------------------------------------------------------

class FiImage
{
public:
    FiImage();
    virtual ~FiImage();

    //! This object is non-copyable and non-copy assignable.
    FiImage(const FiImage& i) = delete;
    FiImage& operator=(const FiImage& i) = delete;

    //! Factory creation method.
    static std::unique_ptr<FiImage> CreateImageObject(IMAGE_FORMAT imageType);

    //! Provide query to determine if the image object has an image attached.
    virtual bool IsImageAttached();

    //! @brief Image header creation method.
    //! @param fourcc   One of the supported FAS_VIDEO @ref fv_fourcc "FourCC" codes.
    //! @param meta     Pointer to the metadata.
    //! @returns The number of bytes the header uses, or a negative number
    //!          indicating an error.
    virtual int CreateHeader(uint32_t fourcc, const pm_metadata *meta) = 0;

    //! Image header access method.
    virtual std::unique_ptr<FAS_VIDEO::HEADER_DATA> GetHeader() = 0;

    //! Attach an image package to this image object.
    virtual FAS_VIDEO::IMAGE_STATUS AttachImagePkg(std::unique_ptr<FAS_VIDEO::FiImagePkg> imagePkgToAttach);

    virtual uint32_t GetAttachedImageSize();
    virtual const uint8_t *GetAttachedImageData();

    virtual const uint8_t *GetAttachedFrameData(uint32_t& attachedFrameSize,
                                                FAS_VIDEO::FRAME_ORIENTATION frameOrientation = FAS_VIDEO::FRAME_ORIENTATION::TOP_DOWN);

    virtual const uint8_t *GetAttachedFrameData(uint32_t& attachedFrameSize,
                                                FAS_VIDEO::FRAME_PLANAR_DATA& planarDataAttached,
                                                FAS_VIDEO::FRAME_ORIENTATION frameOrientation = FAS_VIDEO::FRAME_ORIENTATION::TOP_DOWN);


    //! Retrieves a copy of the frame data associated with any attached image
    //! data minus any header if one exists. No Implementation here yet!
    virtual std::unique_ptr<FAS_VIDEO::FiFrame> CopyAttachedFrameData();

    //! @name Public Property Accessor Functions.
    //!
    //! These are the public functions used to access the properties of this object.
    //! @{

    virtual FAS_VIDEO::IMAGE_FORMAT GetImageFormat() = 0;
    virtual FAS_VIDEO::IMAGE_STATUS GetImageStatus();
    virtual uint32_t GetImageWidth();
    virtual uint32_t GetImageHeight();
    virtual uint32_t GetBitsPerPixel();
    virtual uint32_t GetRowAlign();
    virtual uint32_t GetFullFramesize();

    virtual uint32_t GetFourCC();

    //! @}

protected:
    virtual bool VerifyImageData(const uint8_t *imageDataToVerify, uint32_t imageDataSize) = 0;
    virtual bool GetAttachedFramePlanarData(FRAME_PLANAR_DATA& planarDataToSet);

    // This is protected and non-virtual to manage the "known" private data.
    bool HasKnownImageInfo();

    FAS_VIDEO::IMAGE_STATUS status; //!< The status of this image object.

    //! When FiImage::CreateHeader is called this information is copied and persists.
    //! When image data is attached any specific data to that format is maintained
    //! within its class.
    //!
    //! It is important to understand that this data could be in an empty state if
    //! FiImage::CreateHeader was never invoked.  In that case FiImage::headerFourCC
    //! will be set to FOURCC_NONE
    uint32_t headerFourCC;
    pm_metadata headerMetaData;

    //! Holds the image data associated with this image object.
    std::unique_ptr<FAS_VIDEO::FiImagePkg> imagePkg;

private:
    //! @{
    //!
    //! This data is used to store known image information on the first attachment to
    //! the object.  If at least one attachment has occurred these data fields should
    //! be populated and each additional attachment is expected to conform to the image
    //! information.
    //!
    //! While it can not be assured that there will always be image data
    //! attached to this image object if any image data had been previously
    //! attached this object should be able to rely on this stored "known"
    //! image data to provide framesize and other image information to perform
    //! needed calculations.
    //!
    //! When an image object is reused for a stream each attachment should look
    //! identical to the all image frames in a stream with regard to this information.
    //!
    bool hasKnownImageInfo;
    uint32_t knownImageWidth;
    uint32_t knownImageHeight;
    uint32_t knownBitsPerPixel;
    uint32_t knownRowAlign;

    //! @}

};


} // End of FAS_VIDEO

#endif /* !FASTEC_IMAGE_H_ */
