/*!
 * @file fi_lutfile.h
 * @brief Format of the user LUT file data
 *
 * Copyright 2016 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */
#ifndef _FI_LUTFILE_H_
#define _FI_LUTFILE_H_

#include <stdint.h>

// This depends on #defines in gige_msg.h but we don't include it because
// its path depends on whether it is a FasMotion or camera ltib build.

//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

//! Settings for the lutDataType of each LUT_*FILE data structure.  This setting
//! determines how the lut data is configured in LUT file data.
typedef enum
{
    LUT_SERIES3 = 0,            //!< Series3 Camera LUT file structure.
    LUT_SERIESNC_MONO,          //!< SeriesNC Camera Monochrome Sensor LUT file structure.
    LUT_SERIESNC_COLOR,         //!< SeriesNC Camera Color Sensor LUT file structure.
    LUT_SERIES_UNKNOWN          //!< Unexpected LUT data type.
} LUT_DATA_TYPES;

// Note:  The LUT_FILE structure is used for two types of camera systems.  It maintains
// backward compatibility with the series3 camera systems and it is utilized for monochrome
// seriesNC camera systems.  For the seriesNC camera systems that support RGB LUT tables
// they are stored in the LUT_RGB_FILE data structure.  The entire LUT_FILE structure
// should overlay byte-for-byte over the LUT_RGB_FILE structure from beginning to the end.
// This is a design restriction and it should be maintained for future modifications.

#pragma pack(1)
struct spline_knot {
    uint16_t x;		// 0-1023
    uint16_t y;		// 0-255
};
#pragma pack(0)

#pragma pack(1)
typedef struct LUT_HEADER
{
    // Null-terminated user description for this LUT data.
    char    description[LUT_DESCRIPTION_LEN + 1];

    // Previously an "unused for alignment" byte we are now utilizing it as a lutDataType
    // to determine if 1 channel or 3 channels of data are provided after this byte.
    // The possible values are included in LUT_DATA_TYPES.  The only valid setting for
    // lutDataType in this data structure is LUT_SERIES3 or LUT_SERIESNC_MONO.
    uint8_t    lutDataType;
} LUT_HEADER;
#pragma pack(0)

#pragma pack(1)
typedef struct LUT_FILE
{
    LUT_HEADER hdr;

    // Spline control points, aka knots, are in order by X coordinate, low to high.
    // First has x=0. Last has x=1023. Any entries after x=1023 are ignored.
    struct spline_knot knots[ LUT_MAX_KNOT_COUNT ];

    // Look-up table converts from 10-bit to 8-bit data.
    uint8_t table[ LUT_MAX_TABLE_COUNT ];
}LUT_FILE;
#pragma pack(0)

#pragma pack(1)
typedef struct LUT_RGB_FILE
{
    LUT_HEADER hdr;

    // Each color channel will contain spline_knot record followed by a byte array for the LUT.

    // Spline control points, aka knots, are in order by X coordinate, low to high.
    // First has x=0. Last has x=1023. Any entries after x=1023 are ignored.
    struct spline_knot redKnots[ LUT_MAX_KNOT_COUNT ];
    uint8_t redTable[ LUT_MAX_TABLE_COUNT ];	// Look-up table converts from 10-bit to 8-bit data.

    struct spline_knot greenKnots[ LUT_MAX_KNOT_COUNT ];
    uint8_t greenTable[ LUT_MAX_TABLE_COUNT ];	// Look-up table converts from 10-bit to 8-bit data.

    struct spline_knot blueKnots[ LUT_MAX_KNOT_COUNT ];
    uint8_t blueTable[ LUT_MAX_TABLE_COUNT ];	// Look-up table converts from 10-bit to 8-bit data.

}LUT_RGB_FILE;
#pragma pack(0)

#endif // _FI_LUTFILE_H_

