/*!
 * @file fi_message.h
 * @brief Definitions and declarations for the Message object(s)
 *    used by the Fastec Thread library.
 *
 * Copyright 2019 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef FI_MESSAGE_H_
#define FI_MESSAGE_H_
#pragma once

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <stdint.h>

#include <iostream>
#include <memory>
#include <stack>

#include "fas_thread.h"

namespace FAS_THREAD
{
//--------------------------------------------------------------------
// Constant Definitions
//--------------------------------------------------------------------

//! Global Message request codes.
enum REQ_CODE
{
    NOP,                      //!< NOP

    /*!
     * @brief Shutdown further request queue processing
     *
     * The HALT message provides a means to ensure that all of a Thread's
     * pending messages are processed before the thread itself is destroyed.
     * If a Thread's request queue receives additional input Messages after
     * a HALT has been placed on the request queue, the Thread's messageAdd()
     * function will raise an exception (in the context of the sending Thread).
     *
     * A Thread receiving a HALT could confirm that none of its own Messages
     * are outstanding at the time it processes the HALT Message.
     *
     * Threads receiving HALT must forward the Message (unless the HALT
     * Message belongs to the receiving Thread).
     */
    HALT,

    /*!
     * @brief Dump status
     *
     * This is a generic diagnostic request. Receiving threads are expected
     * to dump their status (however that may be implemented) as human-readable
     * strings to the file descriptor given in the 'stream' field.
     */
    DUMP
};

//! Global Message response codes.
enum RESP_CODE
{
    NONE,       //!< No response provided.
    OK,         //!< All is well.
    FAULT       //!< There was a problem.
};

//--------------------------------------------------------------------
// Forward Declarations
//--------------------------------------------------------------------

class MsgThread;

//--------------------------------------------------------------------
// Class Declaration
//--------------------------------------------------------------------

/*!
 * @brief Base Message class for inter-thread communication
 *
 * Messages may be passed among Threads in whatever means is convenient. Threads
 * may explicitly pass Messages to one another by using the receiving Thread's
 * messageAdd() method.
 *
 * Alternatively, a Message's originator can fill in the list of Threads through
 * which a Message should pass, in turn. The Message's own forward() method will
 * pass the Message to the next Thread's input queue. When the route has been
 * completed (no Threads remain in the route), forward() will return the Message
 * to its owner.
 */
class Message
{
public:
    Message();
    virtual ~Message();

    virtual void forward();

public:
    FAS_THREAD::REQ_CODE req;           //!< Message's request code.
    FAS_THREAD::RESP_CODE resp;         //!< Message's response code.
    FAS_THREAD::MsgThread *owner;       //!< Owning Thread.

    //! Thread objects through which this message is routed.
    //! The route is a LIFO queue, or a stack...
    std::unique_ptr<std::stack<FAS_THREAD::MsgThread *>> route;

    //! Output stream for FAS_THREAD::REQ_CODE::DUMP request.
    std::unique_ptr<std::ostream> stream;
};
} /* namespace FAS_THREAD */
#endif /* FI_MESSAGE_H_ */
