/*!
 * @file fas_stream_mp4.h
 * @brief The h.264 MP4 stream object of the Fastec Video library.
 *
 * Copyright 2019 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef FASTEC_MP4_STREAM_H_
#define FASTEC_MP4_STREAM_H_

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <stdint.h>
#include <memory>
#include <vector>

#include <fi_stream.h>
#include <ffmpeg_container.h>

#include "fi_durations.h"

namespace FAS_MP4
{
//--------------------------------------------------------------------
// Constant Definitions
//--------------------------------------------------------------------

//! @brief List of H.264 encoder preset settings we are supporting.
//! The preset encoder option defines how efficient the codec is at
//! compressing the video frame data.  A very fast setting is less
//! efficient and produces larger file sizes and playback bitrates but
//! produces the video file quickly whereas a very slow setting takes
//! longer to encode the video and is more efficient at compressing
//! the video frames into smaller filesize and faster playback bitrates.
typedef struct H264_PRESETS
{
    FAS_VIDEO::H264_SPEEDS speed;
    std::string speedPreset;
} H264_PRESETS;

//--------------------------------------------------------------------
// Forward Declarations
//--------------------------------------------------------------------

//--------------------------------------------------------------------
// Class Declaration
//--------------------------------------------------------------------

class FiVideoMP4Stream : public FAS_VIDEO::FiVideoStream
{
public:
    FiVideoMP4Stream(uint32_t frameQualityPct = FAS_VIDEO::DEFAULT_H264_QUALITY_PCT,
                     FAS_VIDEO::H264_SPEEDS compressionSpeed = FAS_VIDEO::H264_SPEEDS::H264_MEDIUM,
                     uint32_t keyFrameIntervalToUse = FAS_VIDEO::DEFAULT_H264_KEYFRAME_INTERVAL);

    virtual ~FiVideoMP4Stream() override;

    // This object, for now, is non-copyable and non-copy assignable.
    FiVideoMP4Stream(const FiVideoMP4Stream& i) = delete;
    FiVideoMP4Stream& operator=(const FiVideoMP4Stream& i) = delete;

    FAS_VIDEO::STREAM_STATUS PrepareForWriting(const std::string& videoNameToCreate,
                                               uint32_t frameRateToUse,
                                               uint32_t frameWidthToUse,
                                               uint32_t frameHeightToUse,
                                               uint32_t frameCountToUse) override;

    virtual FAS_VIDEO::STREAM_STATUS SetH264Options(uint32_t frameQualityPct = FAS_VIDEO::DEFAULT_H264_QUALITY_PCT,
                                                    FAS_VIDEO::H264_SPEEDS compressionSpeed = FAS_VIDEO::H264_SPEEDS::H264_MEDIUM,
                                                    uint32_t keyFrameIntervalToUse = FAS_VIDEO::DEFAULT_H264_KEYFRAME_INTERVAL);

    virtual FAS_VIDEO::STREAM_STATUS PrepareForReading(const std::string& videoNameToOpen) override;

    virtual FAS_VIDEO::STREAM_STATUS AddVideoFrame(std::shared_ptr<FAS_VIDEO::FiImage> imageDataFrame) override;
    virtual FAS_VIDEO::STREAM_STATUS AddPlaceholderFrame() override;
    virtual FAS_VIDEO::STREAM_STATUS EditVideoFrame(uint32_t frame_num, std::shared_ptr<FAS_VIDEO::FiImage> imageDataFrame) override;

    virtual FAS_VIDEO::STREAM_STATUS IsStreamWriterComplete() override;

    virtual std::unique_ptr<FAS_VIDEO::FiImage> LoadVideoFrame(uint32_t frameNum) override;

    virtual FAS_VIDEO::STREAM_STATUS CloseStream() override;

    virtual uint32_t GetLastFrameAdded() override;
    virtual uint32_t GetFramesEncodedCount() override;

protected:
    static uint32_t CalculateCRFOption(uint32_t h264QualityPct);
    static uint32_t ValidatePresetOption(FAS_VIDEO::H264_SPEEDS h264Speed);
    static uint32_t SetKeyFrameInterval(uint32_t h264KeyFrameInterval);

    //! FFmpeg Container object responsible for building the output video.
    std::shared_ptr<FI_FFMPEG::FFmpegContainer> ffmpegVideo;

    bool receivedFirstFrame;        //!< Keep track of whether or not a first frame has arrived to stream.

    uint32_t mp4FramesLost;         //!< Keeps track of the number of frames lost.
    uint32_t mp4TotalFramesLost;    //!< Keeps track of the total number of frames lost.

    uint32_t presetValue;
    uint32_t crfValue;
    H264_PRESETS presetsList[4];    //!< "slower", "medium", "fast", "veryfast" entries.
    uint32_t keyFrameInterval;      //!< How often does this stream force a keyframe in an MP4 video.
    bool swOverrideFlag;            //!< If quality/speed settings warrant software override flag set.

    std::shared_ptr<FAS_VIDEO::FiDurations> addVideoFrameStats;
    std::shared_ptr<FAS_VIDEO::FiDurations> addPlaceholderFrameStats;

    bool        FFmpegContainerOK();

    bool        OkToWrite(const char *whoIsAsking, bool addImageAttachCheck);

    uint32_t    ClearMP4FramesLost();
    uint32_t    GetMP4FramesLost();
    uint32_t    GetMP4TotalFramesLost();
    uint32_t    IncrementMP4FramesLost();
};

} // End of FAS_MP4 Namespace

#endif /* !FASTEC_MP4_STREAM_H_ */
