/*!
 * @file  lssm.h
 * @brief Local Streaming Shared Memory interface
 *
 * Copyright 2015 (c) Fastec Imaging Corporation as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of Fastec
 * Imaging Corporation. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement with
 * Fastec Imaging Corporation.
 */

#ifndef _LSSM_H_
#define _LSSM_H_

#include <stdint.h>
#include <time.h>

#ifdef __cplusplus
extern "C" {
#endif

int  lssm_lock(void *p_shared_mem);
int  lssm_timedlock(void *p_shared_mem, struct timespec *tm_ptr);
int  lssm_unlock(void *p_shared_mem);

//! @name Functions for accessing buffer information.
//! @{

//! @name For use only by owner of the buffer (e.g. after get_write_buffer, get_newest, get_oldest).
//! @{
int  lssm_get_mode(void *p_shared_mem);
void lssm_set_mode(void *p_shared_mem, int mode);
int  lssm_get_frame_num(void *p_shared_mem, int bufid);
void lssm_set_frame_num(void *p_shared_mem, int bufid, int frame_num);
uint32_t lssm_get_processed_size(void *p_shared_mem, int bufid);
void lssm_set_processed_size(void *p_shared_mem, int bufid, uint32_t processed_size);
uint64_t lssm_get_image_address(void *p_shared_mem, int bufid);
uint32_t lssm_get_image_header_size(void *p_shared_mem);
void lssm_copyin_image_header(void *p_shared_mem, uint8_t *dst, uint32_t dst_size);
void lssm_copyout_image_header(void *p_shared_mem, uint8_t *src, uint32_t src_size);

//! @name Functions for writer, which is the server.
//! @{
int  lssm_get_write_buffer_l(void *p_shared_mem, int overwrite_oldest);
void lssm_set_write_done_l(void *p_shared_mem, int bufid);
//! @}

//! @name Functions for reader, which is the client.
//! @{
int  lssm_get_newest_l(void *p_shared_mem);
int  lssm_get_oldest_l(void *p_shared_mem);
void lssm_release_buffer_l(void *p_shared_mem, int bufid);
//! @}

//! Get number of non-empty buffers.
int lssm_non_empty_buffers_l(void *p_shared_mem);

//! Check server fault status.
int lssm_get_server_fault_l(void *p_shared_mem);

//! Check client fault status.
int lssm_get_client_fault_l(void *p_shared_mem);
//! @}

//! @name Functions for server only.
//! @{
int  lssm_server_init(void **p_shared_mem, uint64_t *buf_addrs, uint32_t buf_size, unsigned buf_count);
int  lssm_server_reset(void *p_shared_mem);
void lssm_server_cleanup(void *p_shared_mem);
int  lssm_server_enqueue(void *p_shared_mem, uint32_t *message);
int  lssm_server_dequeue(void *p_shared_mem, uint32_t *message);
int  lssm_server_signal_l(void *p_shared_mem);
int  lssm_server_wait_l(void *p_shared_mem, uint64_t timeout_usecs);
int  lssm_server_wait_until_l(void *p_shared_mem, struct timespec *until_time);
int  lssm_server_clear_fault_l(void *p_shared_mem);        //!< @brief Clear fault indication from client
int  lssm_server_set_fault_l(void *p_shared_mem);          //!< @brief Send fault indication to client
//! @}

//! @name Functions for client only.
//! @{
int  lssm_client_init(void **p_shared_mem);
int  lssm_client_cleanup_l(void *p_shared_mem);
int  lssm_client_signal_l(void *p_shared_mem);
int  lssm_client_wait_l(void *p_shared_mem, uint64_t timeout_usecs);
int  lssm_client_clear_fault_l(void *p_shared_mem);        //!< @brief Clear fault indication from server
int  lssm_client_set_fault_l(void *p_shared_mem);          //!< @brief Send fault indication to server
//! @}
//! @}

#ifdef __cplusplus
}
#endif
#endif /* _LSSM_H_ */

