/*!
 * @file fi_hwmon.c
 * @brief Hardware monitor interface for the Fastec Imaging FPGA driver.
 *
 * @sa https://www.kernel.org/doc/html/latest/hwmon/hwmon-kernel-api.html
 *
 * Copyright (C) 2021 Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------

#include <linux/bits.h>
#include <linux/clk.h>
#include <linux/delay.h>
#include <linux/hwmon.h>
#include <linux/io.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/of.h>

#include "fi_pci.h"
#include "fi_fpga_dev.h"

//-----------------------------------------------------------------------------
// Constant Definitions
//-----------------------------------------------------------------------------

//! The name we call ourself.
#define HWMON_NAME      "fi_hwmon"

//! @anchor ChlGroup
//! @name Available Temperature Channels
//! @brief These are the "channels" used to index into the #fi_hwmon_info array.
//! @{

//! The EMC2113 Fan Controller temperature.
#define TEMP_EMC2113_CHANNEL    0

//! The EM2120 Power Module temperature.
#define TEMP_EM2120_CHANNEL 1

//! The internal FPGA temperature.
#define TEMP_FPGA_CHANNEL   2
//
//! @}

// ------------------------------------------------------------------
//! @name EMC2113 Fan Controller Definitions
//! @{
const uint8_t EMC2113_I2C_ADDR = 0x2E;      //!< I2C address of the fan controller on the IO board.

const uint8_t EMC2113_EXT_DIODE1_HIGH = 0x02;
const uint8_t EMC2113_EXT_DIODE1_LOW = 0x03;

const uint8_t EMC2113_PWM_VALUE = 0x40;

const uint8_t EMC2113_TACH_VALUE_HIGH = 0x4E;
const uint8_t EMC2113_TACH_VALUE_LOW = 0x4F;
//
//! @}
// ------------------------------------------------------------------

// ------------------------------------------------------------------
//! @name EM2120 Power Module
//! @{
// EM2120 I2C address.
const uint8_t EM2120_I2C_ADDR = 0x40;

// EM2120 Registers (of interest)
const uint8_t EM2120_STATUS_WORD = 0x79;
const uint8_t EM2120_VOUT = 0x8B;
const uint8_t EM2120_IOUT = 0x8C;
const uint8_t EM2120_TEMP = 0x8E;

// em2120 bit definitions. "Power not good" is bit3
// in the MS byte.
const uint8_t EM2120_POWER_NOT_GOOD = (uint8_t)(1 << 3);
//
//! @}
// ------------------------------------------------------------------

// ------------------------------------------------------------------
//! @name FPGA External Temperature Sensor
//! @{
// TMP451 I2C address.
const uint8_t FPGA_TMP451_I2C_ADDR = 0x4C;

// TMP451 Registers (of interest)
const uint8_t TMP451_REG_R_REMOTE_TEMPH = 0x01;
const uint8_t TMP451_REG_R_REMOTE_TEMPL = 0x10;

//
//! @}
// ------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Variable Declarations
//-----------------------------------------------------------------------------

// ===================================
// TODO:
// - The PWM isn't very interesting until you allow the system to set the PWM.
// - Other temperature sensors on the head should be added.
// ===================================

//! This is the structure that defines all the hardware monitoring information
//! the driver will provide. It is separated into "channels" (@ref ChlGroup "Available Temperature Channels".)
static const struct hwmon_channel_info *fi_hwmon_info[] =
{
    HWMON_CHANNEL_INFO(chip,
                       HWMON_C_REGISTER_TZ | HWMON_C_UPDATE_INTERVAL),

    // Channel 0: EMC2113 Fan Controller temperature.
    // Channel 1: EM2120 temperature
    // Channel 2: FPGA's TMP451 sensor
    HWMON_CHANNEL_INFO(temp,
                       (HWMON_T_INPUT | HWMON_T_LABEL),
                       (HWMON_T_INPUT | HWMON_T_LABEL),
                       (HWMON_T_INPUT | HWMON_T_LABEL)),

    HWMON_CHANNEL_INFO(pwm,
                       HWMON_PWM_INPUT),

    HWMON_CHANNEL_INFO(fan,
                       HWMON_F_INPUT | HWMON_F_LABEL),

    // End of the table.
    NULL
};

/*!
 * @brief Read values from the em2120.
 * @return The register reading in milli-units (1/1000 volts, degrees, etc.)
 */
uint32_t read_em2120_reading(int reg_addr, bool mantissa_only)
{
    uint8_t reg_value[2];
    int exponent;   // 'N'
    int mantissa;   // 'Y'
    uint32_t result;
    extern FI_FPGA_DEV fi_fpga_dev;

//    printk("%s: reg_addr=0x%x, mantissa_only=%d\n", __FUNCTION__, reg_addr, mantissa_only);

    reg_value[0] = reg_value[1] = 0;
    if (fi_i2c_read(&fi_fpga_dev, EM2120_I2C_ADDR, reg_addr, reg_value, FI_I2C_MAIN, 2, 1, false))
    {
        printk(KERN_ERR "%s: Error doing I2C!\n", __FUNCTION__);
        return 0;
    }

    if (mantissa_only)
    {
        // For mantissa only values, the exponent is presumed to be 2^-13.
        exponent = -13;
        mantissa = (((int)reg_value[1]) << 8) + ((int)reg_value[0]);
    }
    else
    {
        // Extract the information.
        exponent = (int)((reg_value[1] >> 3) & 0x1F);
        if ((exponent & 0x10) == 0x10)
        {
            exponent = -16 + (exponent & 0xF);
        }
        mantissa = ((int)(reg_value[1] & 0x7) << 8) + ((int)reg_value[0]);
        if ((mantissa & 0x400) == 0x400)
        {
            mantissa = -1024 + (mantissa & 0x3FF);
        }
    }

    // Our result is ((Y * 2^N) * 1000).
    if (exponent < 0)
    {
        result = (mantissa * 1000) / (1 << (-exponent));
    }
    else
    {
        result = ((mantissa * (1 << exponent)) * 1000);
    }
//    printk("%s: mantissa=%d, exponent=%d, result=%d\n", __FUNCTION__, mantissa, exponent, result);
    return result;
}


/*!
 * @brief Return the EM2120 temperature in millidegrees Celsius.
 */
uint32_t read_em2120_temperature(void)
{
    uint32_t temp_val;

    // Return the em2120 temperature.
    temp_val = read_em2120_reading(EM2120_TEMP, false);
//    printk("%s: temp_val=%d\n", __FUNCTION__, temp_val);
    return temp_val;
}


/*!
 * @brief Return the fan RPM.
 */
uint32_t read_fan_tach(void)
{
    uint8_t tach[2];
    uint32_t fan_tach;
    extern FI_FPGA_DEV fi_fpga_dev;

//    printk("%s\n", __FUNCTION__);

    // Read the high and low bytes for the external diode1. There are 2 data bytes,
    // 1 command byte, and it is not read-only because we have to write the command.
    tach[0] = tach[1] = 0;
    if (fi_i2c_read(&fi_fpga_dev, EMC2113_I2C_ADDR, EMC2113_TACH_VALUE_HIGH, tach, FI_I2C_IO, 2, 1, false))
    {
        printk(KERN_ERR "%s: Error doing I2C!\n", __FUNCTION__);
        return 0;
    }

    // Convert to the tach value. From the datasheet...
    fan_tach = ((((uint32_t)tach[0]) << 5) | ((uint32_t)tach[1] >> 3));
    fan_tach = (3932160 * 2) / fan_tach;


    // All done. Need things in units of millidegrees Celsius.
    return fan_tach;
}

/*!
 * @brief Return the fan PWM.
 */
uint32_t read_fan_pwm(void)
{
    uint8_t pwm_value;
    extern FI_FPGA_DEV fi_fpga_dev;

//    printk("%s\n", __FUNCTION__);

    // Read the high and low bytes for the external diode1.
    pwm_value = 0;
    if (fi_i2c_read(&fi_fpga_dev, EMC2113_I2C_ADDR, EMC2113_PWM_VALUE, &pwm_value, FI_I2C_IO, 1, 1, false))
    {
        printk(KERN_ERR "%s: Error doing I2C!\n", __FUNCTION__);
        return 0;
    }

    // All done. Need things in units of millidegrees Celsius.
    return (uint32_t)pwm_value;
}

/*!
 * @brief Return the fan temperature in millidegrees Celsius.
 */
uint32_t read_fan_temp(void)
{
    uint8_t temp[2];
    int32_t degrees_c;
    extern FI_FPGA_DEV fi_fpga_dev;

//    printk("%s\n", __FUNCTION__);

    // Read the high and low bytes for the external diode1.
    temp[0] = temp[1] = 0;
    if (fi_i2c_read(&fi_fpga_dev, EMC2113_I2C_ADDR, EMC2113_EXT_DIODE1_HIGH, temp, FI_I2C_IO, 2, 1, false))
    {
        printk(KERN_ERR "%s: Error doing I2C!\n", __FUNCTION__);
        return 0;
    }

    // Convert to the full temperature. To get things in millidegrees:
    //    temp = (((T0 * 256) + T1) * 1000) / 256
    degrees_c = (((uint32_t)(temp[0] & 0x7F)) << 8) + ((uint32_t)temp[1]);
    if ((temp[0] & 0x80) == 0x80)
        degrees_c = -degrees_c;

    // All done. Need things in units of millidegrees Celsius.
    return (uint32_t)((degrees_c * 1000) / 256);
}


/*!
 * @brief Read the temperature sensor for the FPGA.
 */
uint32_t read_fpga_temperature(void)
{
    uint8_t oldh, newh, lo_byte;
    uint32_t ret_value;
    extern FI_FPGA_DEV fi_fpga_dev;

//    printk("%s\n", __FUNCTION__);

    /*
     * There is a trick here. We have to read two registers to have the
     * sensor temperature, but we have to beware a conversion could occur
     * between the readings. The datasheet says we should either use
     * the one-shot conversion register, which we don't want to do
     * (disables hardware monitoring) or monitor the busy bit, which is
     * impossible (we can't read the values and monitor that bit at the
     * exact same time). So the solution used here is to read the high
     * byte once, then the low byte, then the high byte again. If the new
     * high byte matches the old one, then we have a valid reading. Else
     * we have to read the low byte again, and now we believe we have a
     * correct reading.
     */
    oldh = lo_byte = newh = 0;
    if (fi_i2c_read(&fi_fpga_dev, FPGA_TMP451_I2C_ADDR, TMP451_REG_R_REMOTE_TEMPH, &oldh, FI_I2C_MAIN, 1, 1, false) ||
        fi_i2c_read(&fi_fpga_dev, FPGA_TMP451_I2C_ADDR, TMP451_REG_R_REMOTE_TEMPL, &lo_byte, FI_I2C_MAIN, 1, 1, false) ||
        fi_i2c_read(&fi_fpga_dev, FPGA_TMP451_I2C_ADDR, TMP451_REG_R_REMOTE_TEMPH, &newh, FI_I2C_MAIN, 1, 1, false))
    {
        printk(KERN_ERR "%s: Error doing I2C!\n", __FUNCTION__);
        return 0;
    }
    if (oldh != newh)
    {
        if (fi_i2c_read(&fi_fpga_dev, FPGA_TMP451_I2C_ADDR, TMP451_REG_R_REMOTE_TEMPL, &lo_byte, FI_I2C_MAIN, 1, 1, false))
        {
            printk(KERN_ERR "%s: Error doing I2C on lo_byte!\n", __FUNCTION__);
            return 0;
        }
    }

    // All is well. The high byte is degrees Celsius, the low byte is the decimal value.
    ret_value = ((uint32_t)newh * 1000) + (((uint32_t)(lo_byte >> 4) * 1000) >> 4);
//    printk("%s: newh=0x%x, lo_byte=0x%x, ret_value=%d\n", __FUNCTION__, newh, lo_byte, ret_value);
    return ret_value;
}

//=======================================================================

/*!
 * @brief The callback used by the hardware monitoring system to read
 * 	  our published sensors.
 *
 * @param data	Pointer to the hardware monitoring device.
 * @param type	The sensor type.
 * @param attr 	Attribute identifier associated with a specific attribute (it's confusing, RTFM).
 * @param channel The sensor channel number (@ref ChlGroup "Available Temperature Channels".)
 * @param val	Pointer to where to put the value.
 *
 * @returns 0 on success, a negative error number otherwise.
 */
static int fi_hwmon_read(struct device *dev,
                         enum hwmon_sensor_types type,
                         u32 attr, int channel, long *val)
{
    int rc;

//    printk("%s: attr=%u, channel=%d\n", __FUNCTION__, attr, channel);

    // Initialize to an error condition.
    rc = -ENOTSUPP;

    switch (type)
    {
    case hwmon_fan:
        if (attr == hwmon_fan_input)
        {
            // Read the tach from the fan.
            *val = read_fan_tach();
            rc = 0;
        }
        break;
    case hwmon_pwm:
        if (attr == hwmon_pwm_input)
        {
            // Read the PWM from the fan.
            *val = read_fan_pwm();
            rc = 0;
        }
        break;
    case hwmon_temp:
        if (attr == hwmon_temp_input)
        {
            // Which channel are we interested in?
            if (channel == TEMP_EMC2113_CHANNEL)
            {
                // Read the temperature from the fan.
                *val = read_fan_temp();
                rc = 0;
            }
            else if (channel == TEMP_EM2120_CHANNEL)
            {
                // Read the temperature from the EM2120.
                *val = read_em2120_temperature();
                rc = 0;
            }
            else if (channel == TEMP_FPGA_CHANNEL)
            {
                // Read the temperature from the FPGA's TMP451.
                *val = read_fpga_temperature();
                rc = 0;
            }
        }
        break;
    default:
        break;
    }

    // Return our result.
    return rc;
}

/*!
 * @brief The callback used by the hardware monitoring system to write
 * 	  to our published sensors (assuming they support being written
 * 	  to).
 *
 * @param data	Pointer to the hardware monitoring device.
 * @param type	The sensor type.
 * @param attr 	Attribute identifier associated with a specific attribute (it's confusing, RTFM).
 * @param channel The sensor channel number (@ref ChlGroup "Available Temperature Channels".)
 * @param val	The value to write.
 *
 * @returns 0 on success, a negative error number otherwise.
 */
static int fi_hwmon_write(struct device *dev,
                          enum hwmon_sensor_types type,
                          u32 attr, int channel, long val)
{
    int rc;

//    printk("%s\n", __FUNCTION__);

    // Initialize to an error condition.
    rc = -ENOTSUPP;

    switch (type)
    {
    case hwmon_pwm:
    default:
        break;
    }

    // Return our result.
    return rc;
}

/*!
 * @brief Called to determine if a particular entry should be visible in the
 *    file system or not.
 *
 * @param data	Pointer to the device private data structure.
 * @param type	The sensor type.
 * @param attr 	Attribute identifier associated with a specific attribute (it's confusing, RTFM).
 * @param channel The sensor channel number (@ref ChlGroup "Available Temperature Channels".)
 *
 * @returns The file mode for this attribute.
 */
umode_t fi_hwmon_is_visible(const void *data, enum hwmon_sensor_types type, u32 attr, int channel)
{
    umode_t rc;

//    printk("%s: type=%d, attr=0x%08x, channel=%d\n", __FUNCTION__, type, attr, channel);

    // Initialize our return value to NOT visible.
    rc = 0;

    // Which sensor are we talking about?
    switch (type)
    {
    case hwmon_fan:
        switch (attr)
        {
        case hwmon_fan_input:
        case hwmon_fan_label:
            rc = 0444;
            break;
        default:
            break;
        }
        break;
    case hwmon_pwm:
        switch (attr)
        {
        case hwmon_pwm_input:
            rc = 0444;
            break;
        default:
            break;
        }
        break;
    case hwmon_temp:
        switch (attr)
        {
        case hwmon_temp_input:
        case hwmon_temp_label:
            rc = 0444;
            break;
        default:
            break;
        }
        break;
    default:
        break;
    }

    // Return the mode.
    return rc;
}

/*!
 * @brief Returns a pointer to a constant string to label the channel information.
 *
 * @param data	Pointer to the device private data structure.
 * @param type	The sensor type.
 * @param attr 	Attribute identifier associated with a specific attribute (it's confusing, RTFM).
 * @param channel The sensor channel number (@ref ChlGroup "Available Temperature Channels".)
 * @param str	Pointer to return the string pointer.
 *
 * @returns 0 on success, a negative error number otherwise.
 */
static int fi_hwmon_read_labels(struct device *dev,
                                enum hwmon_sensor_types type,
                                u32 attr, int channel, const char **str)
{
    int rc;

//    printk("%s: type=%d, attr=0x%08x, channel=%d\n", __FUNCTION__, type, attr, channel);

    // Initialize to an error condition.
    rc = -ENOTSUPP;

    switch (type)
    {
    case hwmon_fan:
        *str = "FiFanRPM";
        rc = 0;
        break;
    case hwmon_pwm:
        *str = "FiFanPWM";
        rc = 0;
        break;
    case hwmon_temp:
        if (channel == TEMP_EMC2113_CHANNEL)
        {
            *str = "FiFanTemp";
            rc = 0;
        }
        else if (channel == TEMP_EM2120_CHANNEL)
        {
            *str = "FiETemp";
            rc = 0;
        }
        else if (channel == TEMP_FPGA_CHANNEL)
        {
            *str = "FiFpgaTemp";
            rc = 0;
        }
        break;
    default:
        break;
    }

    // Return our result.
    return rc;
}

/*
 * @brief The hardware monitor operations structure initialized with our callbacks.
 */
static const struct hwmon_ops fi_hwmon_ops =
{
    .is_visible = fi_hwmon_is_visible,
    .read = fi_hwmon_read,
    .write = fi_hwmon_write,
    .read_string = fi_hwmon_read_labels,
};

/*
 * @brief The top-level data structure for the monitoring driver.
 */
static const struct hwmon_chip_info fi_hwmon_chip_info =
{
    .ops = &fi_hwmon_ops,
    .info = fi_hwmon_info,
};

/*!
 * @brief Set up our hwmon device.
 */
int fi_fpga_setup_hdev(void)
{
    extern FI_FPGA_DEV  fi_fpga_dev;

//    printk("%s\n", __FUNCTION__);

    fi_fpga_dev.hdev = devm_hwmon_device_register_with_info(fi_fpga_dev.fpga_dev,
                                                            HWMON_NAME,
                                                            &fi_fpga_dev,
                                                            &fi_hwmon_chip_info,
                                                            NULL);

    return PTR_ERR_OR_ZERO(fi_fpga_dev.hdev);
}

