/*!
 * @file fi_pci_dev.c
 * @brief Device control code for the PCI device driver.
 *
 * Copyright (C) 2017 Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------

#include <linux/time.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/module.h>
#include <linux/dma-mapping.h>
#include <linux/delay.h>
#include <linux/genalloc.h>
#include <linux/list.h>
#include <linux/memremap.h>
#include <linux/pci.h>
#include <linux/pci-p2pdma.h>
#include <linux/printk.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/atomic.h>
#include <linux/random.h>
#include <linux/unistd.h>

#include "fi_pci_dev.h"
#include "fi_pci.h"

//----------------------------------------------------------------------------
// Variable Delcarations
//----------------------------------------------------------------------------

//! What BAR do we use for I/O?
extern int fi_bar_select;

//! Our private device structure.
FASTEC_PCI_PRIVATE *fi_pci_priv_ptr = NULL;

/*!
 * @brief The function is used to support reference counting of the driver.
 *
 * This is the function which is called when the driver reference count
 * drops to zero. Wake up any task waiting on our completion event. 
 *
 * @param ref	Pointer to the reference structure.
 *
 * @sa https://lwn.net/Articles/557478/
 */
static void fi_pci_percpu_release(struct percpu_ref *ref)
{
    FASTEC_PCI_PRIVATE *priv_ptr = container_of(ref, FASTEC_PCI_PRIVATE, ref);
    printk(KERN_DEBUG "%s", __FUNCTION__);
    complete_all(&priv_ptr->cmp);
}

/*!
 * @brief "Map" the PCIe BARs of our device.
 *
 * During the initialization phase of the driver, enumerates and
 * maps all the PCIe BARs of our device into our IO space so we
 * can access them.
 *
 * In addition, we will setup the Peer-To-Peer bar.
 *
 * @param priv_ptr	Pointer to our driver's data structure.
 * @param dev		Pointer to our PCI device structure.
 */
static void __init map_bars(FASTEC_PCI_PRIVATE *priv_ptr, struct pci_dev *dev)
{
    int i;

    // Loop through all our base address registers and map the
    // memory into our space.
    for (i = 0; i < FASTEC_PCI_BAR_NUM; i++)
    {
        // Grab these entries, because we might need them later.
        priv_ptr->bar_start[i] = pci_resource_start(dev, i);
        priv_ptr->bar_length[i] = pci_resource_len(dev, i);

        // Don't bother with BARs of zero length, or those that start at 0.
        if ((priv_ptr->bar_length[i] == 0) || (priv_ptr->bar_start[i] == 0))
        {
            // Make sure we clear the io_bar value.
            priv_ptr->bar_start[i] = 0;
            priv_ptr->bar_length[i] = 0;
            priv_ptr->io_bar[i] = NULL;
            dev_info(&dev->dev, "BAR[%d] skipped", i);
        }
        else
        {
            // Remap the BAR into our I/O space. Since none of our BARs are
            // prefetchable, this is fine (i.e. don't need write combining).
            priv_ptr->io_bar[i] = pci_iomap(dev, i, 0);
            if (priv_ptr->io_bar[i] == NULL)
            {
                priv_ptr->bar_start[i] = 0;
                priv_ptr->bar_length[i] = 0;
            }
            else
            {
                dev_info(&dev->dev, "BAR[%d] mapped to 0x%p, length %lu", i, priv_ptr->io_bar[i], (long unsigned int)priv_ptr->bar_length[i]);
            }
        }
    }

    // Sanity check, make sure there's something to do here.
    if ((fi_bar_select < FASTEC_PCI_BAR_NUM) && ((priv_ptr->bar_length[fi_bar_select] != 0) && (priv_ptr->bar_start[fi_bar_select] != 0)))
    {
	int rc;

        // Map the specified bar for P2P DMA. Pass a zero length to use the whole bar.
        rc = pci_p2pdma_add_resource(dev, fi_bar_select, 0, 0);
        if (rc)
        {
            dev_warn(&dev->dev, "pci_p2pdma_add_resource() failed\n");
        }
        else
        {
            struct pci_dev *nvme_drv = NULL;

            // Find all NVME drives in the system.
            for (;;)
            {
                // See if we can find any NVME drives in the system.
                nvme_drv = pci_get_class(PCI_CLASS_STORAGE_EXPRESS, nvme_drv);
                if (nvme_drv == NULL)
                {
                    // We are done.
                    break;
                }
                else
                {
                    int dist;

                    // Determine the distance between our driver and the NVME device.
                    // We do this to populate the pci_p2pdma_map_type xarray.
                    dist = pci_p2pdma_distance(dev, &nvme_drv->dev, true);
                    dev_info(&dev->dev, "%s() found [%04x:%04x] dist=%d\n", __FUNCTION__, nvme_drv->vendor, nvme_drv->device, dist);
                }
            }

            // Allocate the resource as p2p memory.
            priv_ptr->virt_bar_addr = pci_alloc_p2pmem(dev, priv_ptr->bar_length[fi_bar_select]);
            if (priv_ptr->virt_bar_addr == NULL)
            {
                dev_warn(&dev->dev, "pci_alloc_p2pmem() failed\n");
            }
        }
    }
    else
    {
        // Give it some error code.
        priv_ptr->virt_bar_addr = NULL;
    }
}

/*!
 * @brief Display the PCI BAR information of our device.
 *
 * @param priv_ptr	Pointer to our driver's data structure.
 * @param dev		Pointer to our PCI device structure.
 */
static void __init dump_bars(FASTEC_PCI_PRIVATE *priv_ptr, struct pci_dev *dev)
{
    int i;

    // Loop through all our base address registers and get the
    // relevant information.
    for (i = 0; i < FASTEC_PCI_BAR_NUM; i++)
    {
        unsigned long bar_start = pci_resource_start(dev, i);
        unsigned long bar_end = pci_resource_end(dev, i);
        unsigned long bar_flags = pci_resource_flags(dev, i);
        unsigned long bar_length = pci_resource_len(dev, i);
        dev_info(&dev->dev, "BAR[%d] 0x%08lx-0x%08lx flags 0x%08lx, length %lu", i, bar_start, bar_end, bar_flags, bar_length);
    }
}

/*!
 * @brief Release all memory associated with the PCI BAR space(s).
 *
 * @param priv_ptr	Pointer to our driver's data structure.
 * @param dev		Pointer to our PCI device structure.
 */
static void unmap_bars(FASTEC_PCI_PRIVATE *priv_ptr, struct pci_dev *dev)
{
    int i;

    // Release our P2P memory.
    if (priv_ptr->virt_bar_addr != NULL)
    {
        pci_free_p2pmem(dev, priv_ptr->virt_bar_addr, priv_ptr->bar_length[fi_bar_select]);
        priv_ptr->virt_bar_addr = NULL;
    }

    // Loop through all our base address registers and release
    // the memory regions.
    for (i = 0; i < FASTEC_PCI_BAR_NUM; i++)
    {
        // Make sure we have something here.
        if (priv_ptr->io_bar[i])
        {
            // Unmap it and remember that it's gone.
            pci_iounmap(dev, priv_ptr->io_bar[i]);
            priv_ptr->io_bar[i] = NULL;
        }
    }
}

/*!
 * @brief This is the probe function for the Fastec PCI driver.
 *
 * This function is called by the PCI core when it has a PCI device that it
 * thinks we might want to control.
 *
 * @param dev   Pointer to the device we might want to take control of.
 * @param id    Pointer to the ID that was used to make this decision.
 * @returns 0 if we want to claim this device, 1 otherwise.
 */
int __init fastec_pci_probe(struct pci_dev *dev, const struct pci_device_id *id)
{
    int rc = 0;

    printk(KERN_DEBUG "%s", __FUNCTION__);

    // Try to enable our device.
    rc = pci_enable_device(dev);
    if (rc)
    {
        dev_err(&dev->dev, "pci_enable_device() failed\n");
        goto err_enable;
    }
    else
    {
        dev_info(&dev->dev, "pci_enable_device() successful");
    }

    // Allocate our private data structure.
    fi_pci_priv_ptr = devm_kzalloc(&dev->dev, sizeof(FASTEC_PCI_PRIVATE), GFP_KERNEL);
    if (fi_pci_priv_ptr == NULL)
        goto err_priv_alloc;

    // Save off a pointer to our device. And do the courtesy back.
    fi_pci_priv_ptr->pci_dev = dev;
    pci_set_drvdata(dev, fi_pci_priv_ptr);

    // Reserve our PCI I/O and memory resources.
    rc = pci_request_regions(dev, MODULE_NAME);
    if (rc)
    {
        dev_err(&dev->dev, "pci_request_regions() failed\n");
        goto err_regions;
    }

    // Allow our device to be a bus master.
    pci_set_master(dev);

#if 0
    // Enable MSI handling for our device.
    rc = pci_enable_msi(dev);
    if (rc)
    {
        dev_info(&dev->dev, "pci_enable_msi() failed\n");
        fi_pci_priv_ptr->msi_enabled = 0;
    }
    else
    {
        dev_info(&dev->dev, "pci_enable_msi() successful\n");
        -fi_pci_priv_ptr > msi_enabled = 1;
    }
#else
    // Disable the MSI handling for our device.
    pci_disable_msi(dev);
    pci_disable_msix(dev);
    fi_pci_priv_ptr->msi_enabled = 0;
#endif

    // Read configuration information from the device that we'll need later.
    pci_read_config_byte(dev, PCI_REVISION_ID, &fi_pci_priv_ptr->revision);
    pci_read_config_byte(dev, PCI_INTERRUPT_PIN, &fi_pci_priv_ptr->irq_pin);
    pci_read_config_byte(dev, PCI_INTERRUPT_LINE, &fi_pci_priv_ptr->irq_line);

    // Initialize our DMA support.
    if (!pci_set_dma_mask(dev, DMA_BIT_MASK(64)))
    {
        pci_set_consistent_dma_mask(dev, DMA_BIT_MASK(64));
        dev_info(&dev->dev, "using a 64-bit dma mask\n");
    }
    else
    {
        dev_info(&dev->dev, "unable to use 64-bit dma mask\n");
        goto err_dma_mask;
    }

    // Initialize the completion object.
    init_completion(&fi_pci_priv_ptr->cmp);

    // Initialize our refcount.
    rc = percpu_ref_init(&fi_pci_priv_ptr->ref, fi_pci_percpu_release, 0, GFP_KERNEL);
    if (rc)
        goto err_dma_mask;

    // Log the information for posterity.
    dev_info(&dev->dev, "irq pin: %d\n", fi_pci_priv_ptr->irq_pin);
    dev_info(&dev->dev, "irq line: %d\n", fi_pci_priv_ptr->irq_line);
    dev_info(&dev->dev, "irq: %d\n", dev->irq);

    // Enable PCI INTx for our device.
    pci_intx(dev, 1);
    {
        u16 cmd_word;

        wmb();
        pci_read_config_word(dev, PCI_COMMAND, &cmd_word);
        dev_info(&dev->dev, "PCI_COMMAND: 0x%04x\n", cmd_word);
    }

    // Request the interrupt for our device.
    rc = request_irq(dev->irq, fi_pci_isr, IRQF_SHARED, MODULE_NAME, (void *)fi_pci_priv_ptr);
    if (rc)
    {
        dev_err(&dev->dev, "Could not request IRQ #%d", dev->irq);
        fi_pci_priv_ptr->irq_line = -1;
        goto err_irq;
    }
    else
    {
        dev_info(&dev->dev, "request irq: %d", dev->irq);
    }


    // Figure out our PCI base address information.
    dump_bars(fi_pci_priv_ptr, dev);
    map_bars(fi_pci_priv_ptr, dev);

    // Get our proc interface setup.
    fi_pci_proc_init(fi_pci_priv_ptr->io_bar[0]);

    // Setup our user space interface.
    fi_pci_cdev_init(fi_pci_priv_ptr);

    // All done.
    return 0;

    // error clean up
err_irq:
    dev_err(&dev->dev, "goto err_irq");
    pci_intx(dev, 0);
err_dma_mask:
    dev_err(&dev->dev, "goto err_dma_mask");
    pci_release_regions(dev);
err_regions:
    dev_err(&dev->dev, "goto err_regions");
    pci_disable_device(dev);
err_enable:
    dev_err(&dev->dev, "goto err_enable");
err_priv_alloc:
    dev_err(&dev->dev, "goto err_priv_alloc");

    // Error exit.
    return rc;
}

/*!
 * @brief This is the function to remove the Fastec PCI driver
 *        from the kernel.
 *
 * @param dev	Pointer to the PCI device structure.
 */
void __exit fastec_pci_remove(struct pci_dev *dev)
{
    FASTEC_PCI_PRIVATE *priv_ptr;

    printk(KERN_DEBUG "%s", __FUNCTION__);

    priv_ptr = pci_get_drvdata(dev);
    if (!priv_ptr)
    {
        printk(KERN_DEBUG "Rather awkward...there is no priv_ptr!");
        return;
    }

    // Close down the user space driver.
    fi_pci_cdev_exit(priv_ptr);

    // Release the proc interface.
    fi_pci_proc_exit();

    // If we had MSI enabled, turn that off.
    if (priv_ptr->msi_enabled)
    {
        pci_disable_msi(dev);
        priv_ptr->msi_enabled = 0;
    }

    // Did we get an IRQ line?
    if (priv_ptr->irq_line >= 0)
    {
        printk(KERN_DEBUG FASTEC_PCI_DRIVER_NAME ": freeing IRQ #%d\n", dev->irq);

        // Make sure we turn off interrupts to our device.
        fi_pci_priv_ptr->irq_line = -1;
        pci_intx(dev, 0);

	// Release the IRQ.
	free_irq(dev->irq, (void *)priv_ptr);
    }

    // Release the BAR memory.
    unmap_bars(priv_ptr, dev);

    // Release the resources associated with the device.
    pci_release_regions(dev);

    // Disable the device.
    pci_disable_device(dev);

    // And we're done.
    printk("Fastec PCI Driver: fastec_pci_remove\n");
}

/*!
 * @brief Return the bar[0] base address.
 *
 * The bar[0] address is the base address for the Fastec FPGA register space.
 */
uint32_t *fi_pci_get_base()
{
    if (fi_pci_priv_ptr == NULL)
        return NULL;

    return (uint32_t *)fi_pci_priv_ptr->io_bar[0];
}
EXPORT_SYMBOL(fi_pci_get_base);

/*!
 * @brief Return the bar[bar_sel] base address.
 *
 * The bar[bar_sel] address is the base address for the Fastec IM FIFO space.
 */
uint32_t *fi_pci_get_im_fifo()
{
    if (fi_pci_priv_ptr == NULL)
        return NULL;

    return (uint32_t *)fi_pci_priv_ptr->io_bar[fi_bar_select];
}
EXPORT_SYMBOL(fi_pci_get_im_fifo);

