/*!
 * @file fi_pci_irq.c
 * @brief IRQ handler for the PCI device driver.
 *
 * Copyright (C) 2017 Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------

#include <linux/init.h>
#include <linux/module.h>
#include <linux/list.h>
#include <linux/printk.h>
#include <linux/slab.h>
#include <linux/atomic.h>
#include <linux/unistd.h>

#include "fi_pci_dev.h"
#include "fi_pci.h"

//----------------------------------------------------------------------------
// Constant Definitions
//----------------------------------------------------------------------------

//! Offset for the interrupt mask register.
#define PCIE_IRQ_MASK           ((u64)0x0038)

//! Offset for the interrupt status register.
#define PCIE_IRQ_STATUS         ((u64)0x003C)

//! Our PCI interrupt callback entry structure.
typedef struct PCI_CB_ENTRY
{
    PCI_CALLBACK cb_ptr;	//!< The pointer to the callback function.
    u32 mask;			//!< The interrupt(s) the callback is interested in.
    struct list_head the_list;	//!< The backpointer to the list of callbacks.
} PCI_CB_ENTRY;

//----------------------------------------------------------------------------
// Variable Delcarations
//----------------------------------------------------------------------------

//! The head of our list.
static LIST_HEAD(pci_cb_list);


/*!
 * @brief Our interrupt service routine.
 */
irqreturn_t fi_pci_isr(int irq, void *dev_id)
{
    uint32_t reg_value;
    uint32_t mask_value;
    PCI_CB_ENTRY *cb_entry = NULL;
    FASTEC_PCI_PRIVATE *priv_ptr = (FASTEC_PCI_PRIVATE *)dev_id;

    // Sanity check. We had better have our private structure!
    if (unlikely(priv_ptr == NULL))
        return IRQ_NONE;

    // In the unlikely event that we have the private structure, but
    // an invalid (-1) irq_line, we need to just return as though
    // we had actually handled it. This is a possibility when we are
    // shutting down.
    if (unlikely(priv_ptr->irq_line < 0))
        return IRQ_HANDLED;

    // Read the status.
    reg_value = ioread32((u32 *)(priv_ptr->io_bar[0] + PCIE_IRQ_STATUS));

    // Read the mask. In FPGA land, a '0' bit enables the interrupt and
    // a '1' bit disables (or masks) the interrupt.
    mask_value = ioread32((u32 *)(priv_ptr->io_bar[0] + PCIE_IRQ_MASK));

    // Disable this interrupt. It's the responsibility of the interrupt handler
    // to re-enable if required.
    iowrite32((reg_value | mask_value), (u32 *)(priv_ptr->io_bar[0] + PCIE_IRQ_MASK));

    // Write the interrupting bits back out to reset the interrupt.
    iowrite32(reg_value, (u32 *)(priv_ptr->io_bar[0] + PCIE_IRQ_STATUS));

    // printk("%s: status=0x%08x mask=0x%08x\n", __FUNCTION__, reg_value, mask_value);

    // Loop through all registered callback functions.
    list_for_each_entry(cb_entry, &pci_cb_list, the_list)
    {
        // Call the registered function if they're interested in this
        // interrupt mask value.
        if (((reg_value & ~mask_value) & cb_entry->mask) != 0)
        {
            (*cb_entry->cb_ptr)(reg_value);
        }
    }

    // pr_info_ratelimited("Got an interrupt!");
    return IRQ_HANDLED;
}

/*!
 * @brief Add a callback function to the list.
 */
void fi_pci_add_callback(PCI_CALLBACK cb, u32 mask)
{
    PCI_CB_ENTRY *entry;

    // Allocate our structure.
    entry = (PCI_CB_ENTRY *)kzalloc(sizeof(PCI_CB_ENTRY), GFP_KERNEL);
    if (entry == NULL)
    {
        printk("%s: Memory alloc failed!\n", __FUNCTION__);
        return;
    }

    // Initialize the structure.
    entry->cb_ptr = cb;
    entry->mask = mask;
    INIT_LIST_HEAD(&entry->the_list);

    // Add it to our list.
    list_add(&entry->the_list, &pci_cb_list);
}
EXPORT_SYMBOL(fi_pci_add_callback);

/*!
 * @brief Remove a callback function from the list.
 */
void fi_pci_remove_callback(PCI_CALLBACK cb)
{
    // Loop through all registered callback functions.
    PCI_CB_ENTRY *cb_entry = NULL;
    PCI_CB_ENTRY *cb_next = NULL;
    list_for_each_entry_safe(cb_entry, cb_next, &pci_cb_list, the_list)
    {
        // Do we have a match?
        if (cb == cb_entry->cb_ptr)
        {
            // Delete this entry.
            list_del(&cb_entry->the_list);

            // All done.
            break;
        }
    }
}
EXPORT_SYMBOL(fi_pci_remove_callback);


